/*
	LPTTEST.C

    Versione 1.2 del 18/07/1996

	Marco Savegnago
    Stradella Ospedale 87
    36100 Vicenza
    msave@tin.it

    Intro

        Fino a qualche anno fa' la porta parallela del Personal Computer era
    considerata, grazie anche al forse troppo semplice design originale di IBM
    una porta di comunicazione unidirezionale (in uscita).
    Gia' da una decina d'anni la ditta statiunitense Travelling Software grazie
    al software LapLink seguita a ruota da altre, ha consentito di usare la
    porta parallela in maniera piu' efficente per trasferire dati ad alta
    velocita' tra computer.
        Il metodo utilizzato in LapLink e' quello della comunicazione
    bidirezionale a 4 bit per volta (nibble) sfruttando le segnalazioni di
    errore o sync per gestire le stampanti.
        La 3A per i propri sistemi a microprocessore programmabile gia' da
    alcuni anni impiega la parallela in modo nibble o byte mode per interrogare
    e configurare ad alta velocita' (in runtime) con un protocollo propietario
    i propri sistemi a microprocessore.
        Da qui la necessita' di una utility per sapere senza problemi se la
    parallela funziona, dov'e' configurata, che interrupt usa per chi magari
    si trova chissa' dove, magari preso con le bombissime, incasinatissimo...

        Aspettando l'USB che e' un bus seriale ad alta velocita' che permette
    velocita' statosferiche per le comuni esigenze la parallela e' piu' che
    sufficente e questa utility puo' essere utile a chi vuol cominciare a fare
    qualche esperimento..

    ... Quindi questo programma verifica il funzionamento di una porta parallela
    mono o bidirezionale EPP (NON ECP!) o del cavo in modo compatibile allo
    standard 3A Zamperla.


    Il programma ritorna i seguenti ERRORLEVEL

	1   Mancanza parametri / Opzione errata
	2   Porta di comunicazione errata
	3   Errore parametri cmdline
	4   Test fallito


     Verifica il funzionamento di una porta parallela mono o bidirezionale
	 USO: LPTTEST <opzioni> <port>
	  port = 1|2|3 (std LPT) 0 = specifica indirizzo es: 03F8
	  opzioni:
		b  - Attiva la porta in modo BYTE
		c  - Esegue il test in maniera continua
		l  - Esegue un test LOOPBACK della porta parallela
		n  - Attiva la porta in modo NIBBLE
		o  - Esegue un test OUTMODE della porta parallela
		q  - Legge lo stato della porta parallela (*)
		s  - Lista le porte presenti nel computer
		t  - Testa la porta di comunicazione
		(*) Solo su parallele monodirezionali

	   Bitmask:
			 Ŀ
			 76543210 Bit
			 
			         
			  9 8 7 6 5 4 3 2  LPTx Pin


	  (*) Nota: La funzionalita' di lettura dello stato della porta
				parallela e' possibile solo su parallele monodirezionali
				(LPT IBM standard)
				Nei moderni computer queste sono quasi sempre configurabili
				anche come EPP o ECP.
				In questo caso e' possibile usare il programma correttamente
				solo con le funzionalita' di scritture assoluta e mai
				relativamnete al valore impostato in precedenza!


    Esempi tipici di impiego del software:

    - Visualizzare le porte di comunicazione parallala presenti nel sistema:

      LPTTEST /s

    - Eseguire un test in modo loop back della porta di comunicazione 1 o
      del cavo ad essa collegato e opera in modo nibble

      LPTTEST /n /l 1

    - Mostra lo stato della porta di comunicazione parallela LPT1

      LPTTEST /q 1

    - Esegue test loopback con opportuno tappo in modo byte mode

      LPTTEST /b /l /t


	PARALLEL PORT   PIN ASSIGNMENT:

	SIGNAL          LEVEL               DB25

	STROBE          N                   1
	DATA BIT 0      P                   2
	DATA BIT 1      P                   3
	DATA BIT 2      P                   4
	DATA BIT 3      P                   5
	DATA BIT 4      P                   6
	DATA BIT 5      P                   7
	DATA BIT 6      P                   8
	DATA BIT 7      P                   9
	ACK             N                   10
	BUSY            P                   11
	PAGE END        P                   12
	SELECT          P                   13
	AUTO FEED       N                   14
	ERROR           N                   15
	INITIAL PRINTER N                   16
	SELECT INPUT    N                   17
	GROUND          -                   18-25



		 Bit Mask    b0  b1  b2  b3  b4  b5  b6  b7
					 |   |   |   |   |   |   |   |
		LPTx  Pin    2   3   4   5   6   7   8   9



	OutTest fa uscire il seguente valore nella porta di comunicazione:

	0xAA, 0x55, 0xA5, 0x5A, 0xFF

	Per eseguire il Loopback test e' necessario ponticellare le uscite della
	porta parallela che si intende impiegare nel seguente sistema:

	Nel caso dell'impiego della porta in nibble mode:

	   1 (Strobe) <-> 10 (Ack)
	   2 (D0)     <-> 11 (Busy)
	   3 (D1)	  <-> 12 (PError)
	   4 (D2)	  <-> 13 (Select)
	   5 (D3)	  <-> 15 (Error)

	Nel caso dell'impiego della porta in byte mode:

    Vedere il circuito da integrare nel tappo di test da applicare alla
    terminazione del cavo

	Per eseguire il Loopback test del cavo parallelo nibble mode e' necessario
	ponticellare le uscite del cavo parallelo lato SYS che si intende
	impiegare nel seguente sistema:

	Nel caso dell'impiego della porta in nibble mode:

	   1 (Strobe) <-> 10 (Ack)
	   2 (D0)     <->  6
	   3 (D1)     <->  7
	   4 (D2)     <->  8
	   5 (D3)     <->  9

    Vedere il circuito da integrare nel tappo di test da applicare alla
    terminazione del cavo

    
    History

    -	Versione 1.0 del 03/07/1996		Inizio
    -	Versione 1.1 del 18/07/1996     Aggiunto help se nibble mode test fallisce
    -	Versione 1.2 del 22/07/1996 	Aggiunto help su byte mode test fallito e show irq
*/

#include	"lpttest.h"

#include	<string.h>
#include	<ctype.h>
#include	<stdio.h>
#include	<dos.h>
#include 	<stdlib.h>
#include	<bios.h>

int _htoi (char *p)
{
	register int i;
	int ret;
	int mult;

	ret=0;
	mult=1;

	i=strlen(p);

	while(i--)
	{
		byte ch = toupper(*(p+i));

		ch -= (isalpha(ch)) ? ('A' - 10) : '0';

		ret += ch * mult;
		mult *= 16;
	}

	return (ret);
}

void uso(void)
{
	 puts
	 (
     "Verifica il funzionamento di una porta parallela mono o bidirezionale e\n"
     "del cavo di connessione secondo standard 3A\n\n"
	 "USO: LPTTEST <opzioni> <port>\n\n"
	 " port = 1|2|3 (std LPT) 0 = specifica indirizzo es: 03F8\n"
	 " opzioni:\n"
	 " b  - Attiva la porta in modo BYTE\n"
	 " c  - Esegue il test in maniera continua\n"
	 " l  - Esegue un test LOOPBACK della porta parallela\n"
	 " n  - Attiva la porta in modo NIBBLE\n"
	 " o  - Esegue un test OUTMODE della porta parallela\n"
	 " q  - Legge lo stato della porta parallela (*)\n"
	 " s  - Lista le porte presenti nel computer\n"
	 " t  - Testa la porta di comunicazione\n"
	 " (*) Solo su parallele monodirezionali"
	 );
}

void    htobin(byte dataIn, byte* dataOut)
{
	int i;

	for(i = 0; i < 8; i++)
		dataOut[i] = (dataIn & (1 << i)) ? 1 : 0;
}

void	SetLPTInputMode(word	port)
{
	byte	in	=	inportb(port + 2);

	outportb(port+2, in | (1 << PC_PAR_LPT_CTR_DIR));
}

void	SetLPTOutputMode(word	port)
{
	byte	in	=	inportb(port + 2);

	outportb(port+2, in & ~(1 << PC_PAR_LPT_CTR_DIR));
}

void    putIn(word comBase, byte dataIn)
{
	byte	dataOut[8];

	htobin(dataIn, dataOut);

	printf("\nPort %Xh Status: %Xh (%d)\n\n", comBase, dataIn, dataIn);
	puts(  "        Ŀ");
	puts(  "Bit     76543210");
	puts(  "        Ĵ");
	printf("Status  %i%i%i%i%i%i%i%i\n",
			dataOut[7], dataOut[6], dataOut[5], dataOut[4],
			dataOut[3], dataOut[2],	dataOut[1],	dataOut[0]);
	puts(  "        ");
	puts(  "                ");
	puts(  "LPT  Pin 9 8 7 6 5 4 3 2\n\n");
}


typedef	struct NOTE
{
	short int pitch;
	short int duration;
};

struct	NOTE notes[] = {{1400, 500}, {1600, 500}, {1200, 500}, {600, 500}, {800, 1000}};

void playnote(struct NOTE note)
{
    outp(0x61, inp(0x61) | 0x03);
    sound(note.pitch);
	delay(note.duration);
    outp(0x61, inp(0x61) & ~0x03);
}


typedef	struct	PARLOOPTESTNIBBLE
{
	word		portOutOffset;
	byte		portOutTestValue;
	byte		portOutPostMask;
	enum bool	invertOut;
	word		portInOffset;
	byte		portInValue;
	enum bool	invertIn;
    const char*	help;
};

struct PARLOOPTESTNIBBLE	parLoopTestNibble[] =
{
{	2,	1<<0, 1<<0, true, 	1, 1<<6, false, "1 (Strobe) <-> 10 (Ack)" },	//	1 (Strobe) <-> 10 (Ack)
{	0,	1<<0, 0<<0,	false,	1, 1<<7, true,  "2 (D0)     <-> 11 (Busy)"},    //  2 (D0)     <-> 11 (Busy)
{	0,  1<<1, 0<<0,	false,	1, 1<<5, false, "3 (D1)     <-> 12 (PError)"},	//	3 (D1)	   <-> 12 (PError)
{	0,  1<<2, 0<<0,	false,	1, 1<<4, false, "4 (D2)     <-> 13 (Select)"},	//	4 (D2)	   <-> 13 (Select)
{	0,  1<<3, 0<<0,	false,	1, 1<<3, false, "5 (D3)     <-> 15 (Error)"},	//	5 (D3)	   <-> 15 (Error)
};

enum bool	LoopBackTestNibble(word portAddress, int index)
{
	byte oldout;
	byte nowin;
	byte nowin2;
	byte nowout;
	enum bool	rc;

	//	Leggo il valore sulla porta

	oldout = inportb(portAddress+parLoopTestNibble[index].portOutOffset);

	//	Devo invertire il valore da impostare in uscita?

	if(parLoopTestNibble[index].invertOut)
		nowout = 	(oldout & (~parLoopTestNibble[index].portOutTestValue));
	else
		nowout =	(oldout | parLoopTestNibble[index].portOutTestValue);

	outportb(portAddress+parLoopTestNibble[index].portOutOffset,nowout);

	//	Quanto vale l'ingresso?

	nowin 	= inportb(portAddress+parLoopTestNibble[index].portInOffset);

	//	Verificalo con quanto dovrebbe valere

	rc = (parLoopTestNibble[index].portInValue&nowin) ? true : false;

	//	Devo negare il risultato?

	if(parLoopTestNibble[index].invertIn)
		rc = rc ? false : true;

	nowout ^=	parLoopTestNibble[index].portOutTestValue;

	outportb(portAddress+parLoopTestNibble[index].portOutOffset,nowout);

	nowin2 = inport(portAddress+parLoopTestNibble[index].portInOffset);

	rc = (nowin2!=nowin) ? true : false;

	//	Rimetti come era l'output

	outportb(portAddress+parLoopTestNibble[index].portOutOffset,oldout | parLoopTestNibble[index].portOutPostMask);

	return rc;
}

void	OutTest(word portAddress, int index)
{
	static const byte outValue[] = { 0xAA, 0x55, 0xA5, 0x5A, 0xFF };

	//	Leggo il valore sulla porta

	outportb(portAddress, outValue[index]);
}

typedef	struct	PARLOOPTESTBYTE
{
	byte		outValue;
    const char*	help;
};

struct PARLOOPTESTBYTE	parLoopTestByte[] =
{
{	0x01,	"2 (D0)" },	//	2 (D0)
{	0x02,	"3 (D1)" }, //	3 (D1)
{	0x04,	"4 (D2)" }, //  4 (D2)
{	0x08,   "5 (D3)" }, //  5 (D3)
{	0x10,	"6 (D4)" },	//	6 (D4)
{	0x20,	"7 (D5)" },	//	7 (D5)
{	0x40,	"8 (D6)" },	//	8 (D6)
{	0x80,	"9 (D7)" }	//	9 (D7)
};

enum bool	LoopBackTestByte(word portAddress, int index, int* n)
{
	byte outStrobe;
	byte nowin;
	enum bool	rc;

    index=index;

	for(*n=0;*n<sizeof(parLoopTestByte) / sizeof(parLoopTestByte[0]);*n++)
	{
		outStrobe	=	inportb(portAddress+2);

		outStrobe	|= (1 << PC_PAR_LPT_CTR_O_STROBE);

		outportb(portAddress+2, outStrobe);

		delay(1);

		SetLPTOutputMode(portAddress);

		outportb(portAddress,parLoopTestByte[*n].outValue);

		delay(1);

		outStrobe	=	inportb(portAddress+2);

		outStrobe	&= ~(1 << PC_PAR_LPT_CTR_O_STROBE);

		outportb(portAddress+2, outStrobe);

		outportb(portAddress, 0);

		SetLPTInputMode(portAddress);

		delay(1);

		//	Quanto vale l'ingresso?

		nowin 	= inportb(portAddress);

		rc = (nowin == parLoopTestByte[*n].outValue) ? true : false;

		if(!rc)
			break;
	}

	return rc;
}

void	outTest(word comBase)
{
	byte	oldValue;
	int		i;

	oldValue = inportb(comBase);

	for(i=0; i < sizeof(notes)/sizeof(notes[0]); ++i)
	{
		OutTest(comBase, i);
		playnote(notes[i]);
	}

	outportb(comBase,oldValue);
}

enum bool 	loopBackTest(word 	portAddress,enum PortMode portMode)
{
	byte	oldValue;
	enum 	bool	loopBackOk;
	int		i;
    int		iErrorNo=0;

	loopBackOk 	= true;
	oldValue 	= inportb(portAddress);

	for(i=0; i < sizeof(notes)/sizeof(notes[0]); ++i)
	{
		if(portMode == NibbleMode)
		{
			if(!LoopBackTestNibble(portAddress, i))
			{
				printf("\a\a\a\aLoopBack Test Nibble su %Xh fallito (%s)!\n", portAddress, parLoopTestNibble[i].help);
				loopBackOk = false;
				break;
			}
		}
		else
		{
			if(!LoopBackTestByte(portAddress, i, &iErrorNo))
			{
				printf("\a\a\a\aLoopBack Test Byte su %Xh fallito (%s)!\n", portAddress, parLoopTestByte[iErrorNo].help);
				loopBackOk = false;
				break;
			}
		}

		playnote(notes[i]);
	}

	outportb(portAddress,oldValue);

	if(loopBackOk)
	{
		delay(200);
		puts("LoopBack Test OK!");
	}

	return loopBackOk;
}

enum bool	isLPTByteMode(word portAddress)
{
	const	byte	test = 0x5A;
	byte	result;
	enum bool	rc;

	result  =	0;
	rc=false;

	SetLPTOutputMode(portAddress);

	outportb(portAddress, 0xA5);

	SetLPTInputMode(portAddress);

	outportb(portAddress, test);

	result  =   inportb(portAddress);

	if(test!=result)
		rc = true;

	SetLPTOutputMode(portAddress);

	return rc;
}

enum	bool	far bIrqFound;
int		far		nInterrupt;

void	(interrupt far *oldVector)(void);

void	interrupt far	LPTIsr(void)
{
	bIrqFound	=	true;

	EOI;

	oldVector();
}

int	LPTIrq(word	wPortBase)
{
	const	IRQVector [] 	=
	{
		IRQ_VECT0,
		IRQ_VECT1,
		IRQ_VECT2,
		IRQ_VECT3,
		IRQ_VECT4,
		IRQ_VECT5,
		IRQ_VECT6,
		IRQ_VECT7
	};

    for (nInterrupt=IRQ_2;nInterrupt<=IRQ_7;nInterrupt++)
    {
        if(nInterrupt!=IRQ_6)
        {
        	disable();

    		oldVector	=	getvect(IRQVector[nInterrupt]);
            setvect(IRQVector[nInterrupt],LPTIsr);

			outportb(wPortBase, 0x80);

			outportb(wPortBase + 2, PC_PAR_LPT_CTR_ENIRQ);

            bIrqFound	=	false;

            SetLPTOutputMode(wPortBase);

            ENABLE_IRQ(nInterrupt);

			EOI;

			enable();

/*
            outportb(wPortBase+ 2,(inportb(wPortBase+ 2) & ~PC_PAR_LPT_CTR_O_STROBE));

            delay(1);

            outportb(wPortBase+ 2,(inportb(wPortBase+ 2) | PC_PAR_LPT_CTR_O_STROBE));

            delay(1);

            outportb(wPortBase+ 2,(inportb(wPortBase+ 2) & ~PC_PAR_LPT_CTR_O_STROBE));
*/
            outportb(wPortBase, 0x00);

//          delay(5);
/*
            outportb(wPortBase, 0x80);

            delay(5);

            outportb(wPortBase, 0x00);

	        delay(10);
*/

            disable();

			DISABLE_IRQ(nInterrupt);

    		setvect(IRQVector[nInterrupt],oldVector);
         	outportb(wPortBase + 2, ~(PC_PAR_LPT_CTR_ENIRQ));

            enable();

            if(bIrqFound)
            {
            	bIrqFound=false;
            	return nInterrupt;
            }
		}
    }

    return -1;
}

void	listLPT(void)
{
	/*
		L'indirizzo delle porte parallele presenti nel PC si trova nel
		segmento 40h agli offset:

		08	-	LPT1
		0A	-	LPT2
		0C	-	LPT3
		0E	-	LPT4	(Solo in !PS2)
	*/

	word __far *    std_prtbase;
	int	i;

	const	 char* modeStr[] = { "NIBBLE", "BYTE  " };

#if __BORLANDC__ >= 0x0452
		std_prtbase = (word __far*) MK_FP( __Seg0040, 8);
#else
#   ifdef _Windows
		std_prtbase = (word __far*) MAKELONG( 8, &_0040H );
#   else
		std_prtbase = (word __far*) MK_FP( 0x0040, 8 );
#	endif
#endif

	puts("\nFound the following LPT\n");

	for(i=0; i<4; i++)
		if(std_prtbase[i])
			printf("LPT%d Address: %Xh %s MODE  IRQ: %i\n", i+1, std_prtbase[i], modeStr[isLPTByteMode(std_prtbase[i])], LPTIrq(std_prtbase[i]));
}


int main(int na, char **a)
{
	word __far *    std_prtbase;
	int 	i;
	int		key;
	enum	bool	done;
	enum	bool	testOk;

	byte	dataIn;
	char* 	p;
	word  	comBase;
	enum 	Action 		action;
	enum    TestMode	testMode;
	enum	PortMode	portMode;

	action		=	None;
	testMode	=	OneShot;
	portMode	=	NibbleMode;
	testOk		=	true;

	comBase=0;

	/*
		L'indirizzo delle porte parallele presenti nel PC si trova nel
		segmento 40h agli offset:

		08	-	LPT1
		0A	-	LPT2
		0C	-	LPT3
		0E	-	LPT4	(Solo in !PS2)
	*/

#if __BORLANDC__ >= 0x0452
		std_prtbase = (word __far*) MK_FP( __Seg0040, 8);
#else
#   ifdef _Windows
		std_prtbase = (word __far*) MAKELONG( 8, &_0040H );
#   else
		std_prtbase = (word __far*) MK_FP( 0x0040, 8 );
#	endif
#endif

    printf("LPTTEST 1.2 - %s Marco Savegnago (3A Zamperla)\n", __DATE__);

/*
**   Controlla che il numero degli argomenti nella linea di comando siano
**   3 altrimenti stampa le istruzioni.
*/
	if (na < 2)
	{
		uso();
		return 1;
	}

	for(i=1;i<na;i++)
	{
		p=a[i];

		if(*p=='-'||*p=='/')
		{
			p++;
			switch(toupper(*p))
			{
				case	'B':
					portMode	=	ByteMode;
					break;

				case	'C':
					testMode	&=  ~(OneShot | Continuosly);
					testMode	|=	Continuosly;
					break;

				case	'L':
					testMode	&=  ~(NibbleMode | OutMode);
					testMode	|=	LoopBack;
					break;

				case	'N':
					portMode	=	NibbleMode;
					break;

				case	'O':
					testMode	&=  ~(NibbleMode | OutMode);
					testMode	|=	OutMode;
					break;

				case	'Q':
					action		=	QueryMask;
					break;

				case	'S':
					action		=	ListPorts;
					break;

				case	'T':
					action		=	TestPorts;
					break;

				default:
					break;
			}
		}
		else
		{
			if(!comBase)
			{
				if(*p < '0' && *p > '3')
				{
					printf("\n\a?? Porta parallela non valida -> %s\n",p);
					uso();

					return 2;
				}
				else
				{
					if(*p=='0')
						comBase=_htoi (p);
					else
						comBase = std_prtbase[(*p - '1')];
				}
			}
			break;
		}
	}

	if((comBase == 0 || action==None) && action!=ListPorts)
	{
		uso();
		return 3;
	}

	done=false;

	while(!done)
	{
		/*	Legge il valore della porta */

		switch(action)
		{
			case	ListPorts:
				listLPT();
				break;

			case	QueryMask:
				if(portMode	== ByteMode)
					SetLPTInputMode(comBase);

				dataIn =	inportb(comBase);

				if(portMode	== ByteMode)
					SetLPTOutputMode(comBase);

				putIn(comBase, dataIn);
				break;

			case	TestPorts:
				if(testMode	& OutMode)
					outTest(comBase);
				else
					testOk = loopBackTest(comBase,portMode);

				break;

			default:
				break;
		}

		if(testMode & OneShot)
			done = true;
		else
		{
			while (bioskey(1))
				bioskey(0);

			puts("\n\aPremi un tasto per continuare Q o ESC per terminare");

			while (bioskey(1) == 0);

			key = bioskey(0) & 0xff;

			if(toupper(key) == 'Q' || key == '\x1b')
			{
				done = true;
				break;
			}

		}
	}

	if(action== TestPorts && testOk == false)
		return 4;

	return 4;
}

